SystemOrganization addCategory: #'_SWE_OOAD'!

AltCardTableMorph subclass: #SWEFreeCell
	instanceVariableNames: 'deck spares piles columns'
	classVariableNames: 'NumColumns NumPiles NumSpares'
	poolDictionaries: ''
	category: '_SWE_OOAD'!

SWEFreeCell class
	instanceVariableNames: 'NumSpares NumPiles NumColumns'!

SWEFreeCell class
	instanceVariableNames: 'NumSpares NumPiles NumColumns'!

!SWEFreeCell class methodsFor: 'parts bin' stamp: 'ChristopherMiller 2/23/2013 19:26'!
descriptionForPartsBin
	^ self
		partName: 'AltFreeCell'
		categories: #('Games' )
		documentation: 'Another FreeCell solitaire card game'! !

!SWEFreeCell class methodsFor: 'class initialization' stamp: 'ChristopherMiller 2/23/2013 19:26'!
includeInNewMorphMenu
	^ true! !

!SWEFreeCell class methodsFor: 'class initialization' stamp: 'ChristopherMiller 2/23/2013 19:26'!
initialize

	NumColumns := 8.
	NumSpares := 4.
	NumPiles := 4.
	^self! !

!SWEFreeCell class methodsFor: 'instance creation' stamp: 'ChristopherMiller 2/23/2013 19:26'!
newGame
	"Answer a new instance"
	"AltFreeCell newGame openInWorld."
	^ self new initialize "openInWindowLabeled: 'FreeCell' inWorld: self currentWorld"! !

!SWEFreeCell class methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
numColumns
	^NumColumns! !

!SWEFreeCell class methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
numPiles
	^NumPiles! !

!SWEFreeCell class methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
numSpares
	^NumSpares! !

!SWEFreeCell methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
column: columnMorph wantsDropOf: aCardOrColumnMorph 
	"Answer true if Column is empty or if (lowest)
	card is opposit color and one higher in face value."
	
	aCardOrColumnMorph isAltCardMorph
		ifTrue: [| topCard | 
			topCard := columnMorph topCard.
			topCard isNil
			ifTrue: [^true]
			ifFalse: [ ^ (aCardOrColumnMorph hasFaceOneLessThan: topCard)
					& (topCard hasDifferentColorThan: aCardOrColumnMorph)]]
		ifFalse: [
			Error signal: '@@FIXME: implement for column drops!!'.
			^ false]! !

!SWEFreeCell methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
columns
	"Private -- Answer my columns"

	^columns! !

!SWEFreeCell methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
container: aContainerMorph wantsDropOf: aCardOrColumnMorph 
	"Answer true if legal drop by FreeCell game rules"
	| dropObjectClass |
	dropObjectClass := aCardOrColumnMorph class.
	(dropObjectClass = AltCardColumnMorph) | (dropObjectClass = AltCardMorph)
	ifFalse: [ ^false ]
	ifTrue: [ aContainerMorph class = AltCardSpareMorph
		ifTrue: [^ self spare: aContainerMorph wantsDropOf: aCardOrColumnMorph]
		ifFalse: [aContainerMorph class = AltCardPileMorph
				ifTrue: [^ self pile: aContainerMorph wantsDropOf: aCardOrColumnMorph]
				ifFalse: [aContainerMorph class = AltCardColumnMorph
						ifTrue: [^ self column: aContainerMorph wantsDropOf: aCardOrColumnMorph]
						ifFalse: [^ false]]]]! !

!SWEFreeCell methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:26'!
dealCards
	"Deal all cards onto the table by adding them to Card Columns."

	self deck shuffle; shuffle; shuffle; shuffle; shuffle; shuffle; shuffle.
	0 to: 51
		do: [ :n |  
			(columns at: (n rem: AltFreeCell numColumns) + 1) add: (deck dealACard) ]
			! !

!SWEFreeCell methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
deck
	"Private -- Answer my deck"

	^deck! !

!SWEFreeCell methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
dropComplete
	"Check to see if any cards can be auto-moved to piles.
	Keep moving cards up if legal unless the lowest face value
	of a Card in any Pile is significantly less than the others
	as noted by #maxFaceNumberToMoveToPiles."
	
	self inCardMove 
		"already doing this"
		ifTrue: [ ^false ]
		ifFalse: 
			[ | goAgain |
				self inCardMove: true.
			 	goAgain := true.
				[ goAgain ] whileTrue:
					[ | maxFaceToMove |
				      goAgain := false.
					 maxFaceToMove := self maxFaceNumberToMoveToPiles.
					 self spares 
						do: [ :container |
							 (self possiblyAnimateCardsFrom: container 
								   ifFaceLessOrEqualTo: maxFaceToMove)
							 ifTrue: [ goAgain := true. ].
						].
			  		 self columns
						do: [ :container |
							  (self possiblyAnimateCardsFrom: container 
								   ifFaceLessOrEqualTo: maxFaceToMove) 
							  ifTrue: [ goAgain := true. ].
						]
				].
				self inCardMove: false.
				self gameIsWon
				ifTrue: [ self showWinnerText ]
			] ! !

!SWEFreeCell methodsFor: 'testing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
gameIsWon
	"Answer true if all cards are in the piles"

	^(self piles inject: 0
			    into: [ :count :pile | count + (pile submorphs size) ]) == 52! !

!SWEFreeCell methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:26'!
help

	self helpWithLabel: 'FreeCell Rules' translated
! !

!SWEFreeCell methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
helpText
	^ '
	All cards are dealt into 8 columns.
 
	Cards in columns build up by alternating color and increasing sequence.

	You can move such "runs" up to the number of free cells plus 1.

	Free cells are the spares and columns which contain no cards.

	There are 4 spares. Each spare can hold a single card.

	The piles build up from the aces in the same suit.' translated! !

!SWEFreeCell methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:26'!
initialize
	"Setup my layout.  Across the top are 4 Spares and 4 Piles,
	then a block of 8 Columns, then a Button area."

	super initialize.
	deck := AltCardDeck new initialize.
	self name: 'FreeCell';
		layoutPolicy: TableLayout new;
		listDirection: #leftToRight;
 		wrapDirection: #topToBottom;
		listCentering: #topLeft;
		makeButtonArea;
		makeColumnArea;
		makePilesArea;
		makeSparesArea.
	self dealCards;
		yourself! !

!SWEFreeCell methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:26'!
makeColumnArea
	"I make new Column Area as a submorph"

	| columnArea |
	columnArea := 
		Morph new initialize; 
			color: Color transparent;
			"height: 460;"
			width: (self width);
			layoutPolicy: TableLayout new;
			layoutInset: 14;
 			listDirection: #leftToRight;
			listCentering: #justified;
			cellPositioning: #topCenter;
			vResizing: #spaceFill.
	1 to: AltFreeCell numColumns 
		do: [ :n |
			| colMorph |
			colMorph := AltCardColumnMorph new initialize.
			colMorph height: 400;
					  borderColor: Color transparent.
			columnArea addMorph: colMorph ].
	columns := columnArea submorphs.
	self addMorph: columnArea! !

!SWEFreeCell methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:26'!
makePilesArea
	"I make new Piles Area as a submorph"

	| pilesArea |
	pilesArea := 
		Morph new initialize; 
			color: Color transparent;
			height: 95;
			width: (self width - 2) / 2;
			layoutPolicy: TableLayout new;
			layoutInset: 1@5;
			cellInset: 9@11;
			listCentering: #center;
 			listDirection: #leftToRight;
			cellPositioning: #topCenter.
	1 to: AltFreeCell numPiles do: [ :n | 
		| pile |
		pile := AltCardPileMorph new initialize;
					setBalloonText: 'Build up a suit in order from the Ace.'.
		pilesArea addMorph: pile ].
	piles := pilesArea submorphs.
	self addMorph: pilesArea! !

!SWEFreeCell methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:26'!
makeSparesArea
	"I make new Piles Area as a submorph"

	| sparesArea |
	sparesArea := 
		Morph new initialize; 
			color: Color transparent;
			height: 95;
			width: (self width - 2) / 2;
			layoutPolicy: TableLayout new;
			layoutInset: 1@5;
			cellInset: 9@11;
			listCentering: #center;
 			listDirection: #leftToRight;
			cellPositioning: #topCenter.
	1 to: AltFreeCell numSpares 
		do: [ :n | 
			| spare |
			spare := AltCardSpareMorph new initialize; 
				setBalloonText: 'A Spare can hold 1 card'.
			sparesArea addMorph: spare ].
	spares := sparesArea submorphs.
	self addMorph: sparesArea! !

!SWEFreeCell methodsFor: 'testing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
maxFaceNumberToMoveToPiles
	"Answer N plus the minimum Card faceNumber"

	| min |

	min := 13. "big"
	self piles 
		do: [ :pile |
			pile isEmpty 
			ifTrue: [ min := min min: 0 ] 
			ifFalse: [ min := min min: (pile submorphs at: 1) faceNumber ]
		    ].
	"N is 2"
	^min + 2! !

!SWEFreeCell methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:26'!
newGame
	"Empty out my Card Containers, shuffle and re-deal."
	
	self removeWinnerText;
		makeUndoStackEmpty;
		removeCards;
	    "Note: (self deck shuffle) is done in (self dealCards)."
		dealCards! !

!SWEFreeCell methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
numberOfFreeCells
	"Answer the number of empty Spares and Columns."

	| numFree |
	numFree := 0.
	self spares   do: [ :spare | spare isEmpty ifTrue: [numFree := numFree + 1]].
	self columns do: [ :spare | spare isEmpty ifTrue: [numFree := numFree + 1]].
	^numFree! !

!SWEFreeCell methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
pickLegalFor: aCard 
	"Answer true if picking up a card from its container is a legal move."
	| cardContainer |
	cardContainer := aCard owner.
	self moveStart: cardContainer. "for Undo"
	(cardContainer isKindOf: AltCardContainerMorph)
		ifFalse: [^ false].
	cardContainer class = AltCardSpareMorph
		ifTrue: [^ true].
	cardContainer class = AltCardPileMorph
		ifTrue: [^ false].
	"else is a Column"
	^ self pickLegalFor: aCard fromColumn: cardContainer! !

!SWEFreeCell methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
pickLegalFor: aCard fromColumn: aColumn
	"Answer true if card is on top or is part of a 'run' and the size of
	the run is less than the number of free cells."

	(self runStartsFrom: aCard in: aColumn) ifFalse: [ ^false ].
	^(self numberOfFreeCells + 1) >= (aColumn numberOfCardsAbove: aCard)! !

!SWEFreeCell methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
pile: pileMorph wantsDropOf: aCardOrColumnMorph 
	"Answer true if Pile is empty and Card is Ace or if  
	card is same suit and one higher in face value."
	
	| card |
	card := aCardOrColumnMorph class = AltCardMorph
				ifTrue: [aCardOrColumnMorph]
				ifFalse: [aCardOrColumnMorph singletonCard].
	card isNil
		ifTrue: [ ^nil ]
		ifFalse: [ pileMorph isEmpty
			ifTrue: [ ^ card face == #Ace ]
			ifFalse: [ | topCard | 
			topCard := pileMorph topCard.
			^ (topCard hasFaceOneLessThan: card)  & (topCard hasSameSuitAs: card)]]! !

!SWEFreeCell methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
piles
	"Private -- Answer my piles"

	^piles! !

!SWEFreeCell methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:26'!
quit
	"@@ToDo: popup dialog Do you REALLY want to Quit?"
	
	super quit! !

!SWEFreeCell methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:26'!
reDealCards
	"Deal all cards onto the table withot shuffling (for game replay)."

	"Assumes cards have been removed first."
	0 to: 51
		do: [ :n |  
			(columns at: (n rem: AltFreeCell numColumns) + 1) add: (deck dealACard) ]
			! !

!SWEFreeCell methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:26'!
removeCards
	"Empty out my Card Containers."
	
	self piles do: [ :container | container makeEmpty ].
	self spares do: [ :container | container makeEmpty ].
	self columns do: [ :container | container makeEmpty ]! !

!SWEFreeCell methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:26'!
restoreGame
	"Put up a dialog to reaf game state from a file.
	See #restoreGameFromDataStream:"

	self restoreGame: 'fre' userMessage: 'Select a FreeCell Game File...'! !

!SWEFreeCell methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:26'!
restoreGameFromDataStream: aDataStream
	"Given an open DataStream, read and restore game state."
	
	"Game state consists of information to reconsitiute the deck, spares, piles, and columns."
	"NYI: undo stack save/restore."
	| stateArray |
	stateArray := aDataStream next.
	(stateArray at: 1) == #AltFreeCell
	ifFalse: [ Error signal: 'Not a saved FreeCell Game!!' ]
	ifTrue: [
		| columnInfo pileInfo spareInfo undoInfo |
		deck := AltCardDeck new initializeFromArray: (stateArray at: 3).
		columnInfo := stateArray at: 5.
		pileInfo     := stateArray at: 7.
		spareInfo   := stateArray at: 9.
		undoInfo    := stateArray at: 11.
		
		1 to: (AltFreeCell numColumns) 
			do: [ :idx |
				(self columns at: idx) restoreFromArray: (columnInfo at: idx) givenDeck: self deck ].
		
		1 to: (AltFreeCell numSpares) 
			do: [ :idx |
				(self spares at: idx) restoreFromArray: (spareInfo at: idx) givenDeck: self deck ].
			
		1 to: (AltFreeCell numPiles) 
			do: [ :idx |
				(self piles at: idx) restoreFromArray: (pileInfo at: idx) givenDeck: self deck ].
			
		undoInfo isNil
		ifTrue: [ self makeUndoStackEmpty ]
		ifFalse: [ Error signal: 'FIXME: Undo save/restore NYI' ].
	]! !

!SWEFreeCell methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:26'!
sameGame
	"Empty out my Card Containers and deal the original deal."
	
	self deck startOver.
	self removeWinnerText;
		makeUndoStackEmpty;
		removeCards;
	    "Note: (self deck shuffle) is done in (self dealCards)."
		reDealCards! !

!SWEFreeCell methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:26'!
saveGame
	"Show a file dialog to the user to save a game.
	See #saveGameToDataStream: for actual data save."
	| sillyDialog |
	self gameIsWon
	ifTrue: [  
		sillyDialog := DenyDialogWindow new.
		sillyDialog text: ('Sorry, I can''t save a game you finished playing' asText).
		self openModal: sillyDialog]
	ifFalse: [self saveGame: 'fre' userMessage: 'Enter File Name to save FreeCell Game...']! !

!SWEFreeCell methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:26'!
saveGameToDataStream: aDataStream
	"Given an open DataStream, save the current game state."
	
	"Game state consists of information to reconsitiute the deck, spares, piles, and columns."
	"NYI: undo stack save/restore."
	| someData |
	someData := 	 { #AltFreeCell. 
							#CardDeck.
							self deck saveStateToArray.
							#Columns.
							self columns 
								collect: [ :container | container saveStateToArray].
							#Piles.
							self piles 
								collect: [ :container | container saveStateToArray].
							#Spares.
							self spares 
								collect: [ :container | container saveStateToArray].
							#Undo.
							nil. "FIXME"
						}.
	aDataStream nextPut: someData.
! !

!SWEFreeCell methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
spare: spareMorph wantsDropOf: aCardOrColumnMorph 
	"Answer true if Spare is empty and Card is singleton."
	| card |
	card := aCardOrColumnMorph class = AltCardMorph
				ifTrue: [aCardOrColumnMorph]
				ifFalse: [aCardOrColumnMorph singletonCard].
	spareMorph isEmpty
		ifTrue: [^ card isNil not]
		ifFalse: [^ false]! !

!SWEFreeCell methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:26'!
spares
	"Private -- Answer my spares"

	^spares! !

AltCardTableMorph subclass: #SWEKlondike
	instanceVariableNames: 'cardDeck deck discards piles columns'
	classVariableNames: 'NumColumns NumPiles'
	poolDictionaries: ''
	category: '_SWE_OOAD'!

!SWEKlondike class methodsFor: 'parts bin' stamp: 'ChristopherMiller 2/23/2013 19:33'!
descriptionForPartsBin
	^ self
		partName: 'Klondike'
		categories: #('Games' )
		documentation: 'Klondike Solitaire card game'! !

!SWEKlondike class methodsFor: 'class initialization' stamp: 'ChristopherMiller 2/23/2013 19:33'!
includeInNewMorphMenu

	^true! !

!SWEKlondike class methodsFor: 'class initialization' stamp: 'ChristopherMiller 2/23/2013 19:33'!
initialize

	NumColumns := 7.
	NumPiles := 4.
	^self! !

!SWEKlondike class methodsFor: 'instance creation' stamp: 'ChristopherMiller 2/23/2013 19:33'!
newGame
	"Answer a new instance"
	"Klondike newGame openInWorld."
	^ self new initialize "openInWindowLabeled: 'Klondike' inWorld: self currentWorld"! !

!SWEKlondike class methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
numColumns
	^NumColumns! !

!SWEKlondike class methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
numPiles
	^NumPiles! !

!SWEKlondike methodsFor: 'auto-move' stamp: 'ChristopherMiller 2/23/2013 19:33'!
animateMoveFrom: aContainer to: otherContainer moving: someCard
	"Do the animation.  If inUndo, do fixup as required."

	| otherSizeAtStart |
	otherSizeAtStart := otherContainer submorphs size.
	super animateMoveFrom: aContainer 
			to: otherContainer 
			moving: someCard.
	inUndo "fixup"
	ifTrue: [ 
		(aContainer isKindOf: AltCardColumnMorph)
			ifTrue: [ aContainer showTopCard ].
		(aContainer isKindOf: AltCardDiscardMorph)
			ifTrue: [ aContainer showTopCards ].
		"@@FIXME: only if were hidden before"
		(otherContainer isKindOf: AltCardColumnMorph)
			ifTrue: [ | cards numCardsMoved | 
				cards := otherContainer submorphs.
				numCardsMoved := (cards size) - otherSizeAtStart.
				(cards size = (numCardsMoved + 1)) 
				ifTrue: [ (cards at: (numCardsMoved + 1)) makeFaceDown. ^true ].
				(cards size > (numCardsMoved + 1))
				ifTrue: [ ((cards at: (numCardsMoved + 2)) isFaceUp)
						ifFalse: [ (cards at: (numCardsMoved + 1)) makeFaceDown ]
				]
			]
	]! !

!SWEKlondike methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
cardDeck

	^cardDeck! !

!SWEKlondike methodsFor: 'event handling' stamp: 'ChristopherMiller 2/23/2013 19:33'!
click: evt from: deckMorph
	"Handle click events from AltCardDeckMorph."
	
	(deckMorph isKindOf: AltCardDeckMorph)
	ifFalse: [ ^false ].

	"ifEmpty, reclaim Cards from Discards, else deal 3 Cards to Discards."
	deckMorph isEmpty
	ifFalse: [ | bottomCard |
			bottomCard := deckMorph topCard. "bottom in undo"
			1 to: (3 min: deckMorph numCards) 
				do: [ :ignored | discards add: deckMorph takeTopCard ].
				
			self pushUndoObject: (AltUndoAction from: discards
												to: deckMorph
												moving: bottomCard)
	]
	ifTrue: [
		[ discards isEmpty ]
			whileFalse: [ deck add: discards takeTopCard ].
		self pushUndoObject: (AltUndoAction from: deckMorph 
												to: discards
												moving: nil "special case")
 	]

	! !

!SWEKlondike methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
column: columnMorph wantsDropOf: aCardOrColumnMorph 
	"Answer true if Column is empty and Card is a King or if (lowest)
	card is opposit color and one higher in face value."
	
	aCardOrColumnMorph isAltCardMorph
		ifTrue: [| topCard | 
			topCard := columnMorph topCard.
			topCard isNil
			ifTrue: [^ aCardOrColumnMorph face = #King ]
			ifFalse: [ ^ (aCardOrColumnMorph hasFaceOneLessThan: topCard)
					& (topCard hasDifferentColorThan: aCardOrColumnMorph)]]
		ifFalse: [
			Error signal: '@@FIXME: implement for column drops!!'.
			^ false]! !

!SWEKlondike methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
columns
	"Private -- Answer my columns"

	^columns! !

!SWEKlondike methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
container: aContainerMorph wantsDropOf: aCardOrColumnMorph 
	"Answer true if legal drop by FreeCell game rules"
	| dropObjectClass |
	dropObjectClass := aCardOrColumnMorph class.
	(dropObjectClass = AltCardColumnMorph) | (dropObjectClass = AltCardMorph)
		ifFalse: [ ^false ].
	aContainerMorph class = AltCardDiscardMorph
		ifTrue: [^ false ].
	aContainerMorph class = AltCardPileMorph
		ifTrue: [^ self pile: aContainerMorph 
					   wantsDropOf: aCardOrColumnMorph ].
	aContainerMorph class = AltCardColumnMorph
		ifTrue: [^ self column: aContainerMorph 
					   wantsDropOf: aCardOrColumnMorph]
		ifFalse: [^ false]! !

!SWEKlondike methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:33'!
dealCards
	"Shuffle cards and deal onto the table."

	self cardDeck shuffle; shuffle; shuffle; shuffle; shuffle; shuffle; shuffle.
	self reDealCards
! !

!SWEKlondike methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
deck
	"Private -- Answer my deck"

	^deck! !

!SWEKlondike methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
discard
	"Private -- Answer my discards"

	^discards! !

!SWEKlondike methodsFor: 'auto-move' stamp: 'ChristopherMiller 2/23/2013 19:33'!
dropComplete
	"Show top Card in Columns & Discard; check for automove or a winner."
	
	| goAgain |
	self inCardMove ifTrue: [ ^false ]. "Already doing this?"
	self columns do: [ :col | col showTopCard ].
	self discard showTopCards.
	self inCardMove: true.

	goAgain := true.
	[ goAgain ]
	whileTrue: [ | maxFaceToMove |
		goAgain := false.
		maxFaceToMove := self maxFaceNumberToMoveToPiles.
		self columns
			do: [ :column |
			 	(self possiblyAnimateCardsFrom: column 
						ifFaceLessOrEqualTo: maxFaceToMove) 
							 ifTrue: [ column showTopCard.
									  goAgain := true. ].
			].
		(self possiblyAnimateCardsFrom: self discard
						ifFaceLessOrEqualTo: maxFaceToMove) 
							 ifTrue: [ goAgain := true. ].
	].
	self inCardMove: false.
	self gameIsWon ifTrue: [ self showWinnerText ]! !

!SWEKlondike methodsFor: 'testing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
gameIsWon
	"Answer true if all cards are in the piles"

	^(self piles inject: 0
			    into: [ :count :pile | count + (pile submorphs size) ]) == 52! !

!SWEKlondike methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:33'!
help

	self helpWithLabel: 'Klondike Rules' translated! !

!SWEKlondike methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
helpText
	^ '
	Cards are dealt into 7 columns, top face up.
 
	Cards in columns build up by alternating color and increasing sequence.

	You can move such "runs" from one column to another to "grow" the run.

	Clicking on the deck deals 3 cards face up to the discards pile.

	You may place a King on an empty column.

	The piles build up from the Aces in the same suit.' translated! !

!SWEKlondike methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:33'!
initialize
	"Setup my layout.  Across the top are Deck Discards and 4 Piles,
	then a block of 7 Columns, then a Button area."

	super initialize.
	self name: 'Klondike';
		layoutPolicy: TableLayout new;
		listDirection: #leftToRight;
 		wrapDirection: #topToBottom;
		listCentering: #topLeft;
		makeButtonArea;
		makeColumnArea;
		makePilesArea;
		makeDeckArea.
	self dealCards;
		yourself! !

!SWEKlondike methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:33'!
makeColumnArea
	"I make new Column Area as a submorph"

	| columnArea |
	columnArea := 
		Morph new initialize; 
			color: Color transparent;
			"height: 460;"
			width: (self width);
			layoutPolicy: TableLayout new;
			layoutInset: 14;
 			listDirection: #leftToRight;
			listCentering: #justified;
			cellPositioning: #topCenter;
			vResizing: #spaceFill.
	1 to: Klondike numColumns 
		do: [ :n |
			| colMorph |
			colMorph := AltCardColumnMorph new initialize.
			colMorph height: 400;
					  borderColor: Color transparent.
			columnArea addMorph: colMorph ].
	columns := columnArea submorphs.
	self addMorph: columnArea! !

!SWEKlondike methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:33'!
makeDeckArea
	"I make new Deck+Discards Area as a submorph"

	| deckArea |
	deckArea := 
		Morph new initialize; 
			color: Color transparent;
			height: 95;
			width: (self width - 2) / 2;
			layoutPolicy: TableLayout new;
			layoutInset: 1@5;
			cellInset: 9@11;
			listCentering: #center;
 			listDirection: #leftToRight;
			cellPositioning: #topCenter.
	deck := AltCardDeckMorph new initialize;
					setBalloonText: 'Tap top card to discard 3 or redeal.'.
	cardDeck := AltCardDeck new initialize.
	deck addCards: cardDeck.
	discards := AltCardDiscardMorph new initialize;
					numCardsToShow: 3;
					setBalloonText: 'Play top Discard.'.
	deckArea addMorph: discards; addMorph: deck.
	self addMorph: deckArea! !

!SWEKlondike methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:33'!
makePilesArea
	"I make new Piles Area as a submorph"

	| pilesArea |
	pilesArea := 
		Morph new initialize; 
			color: Color transparent;
			height: 95;
			width: (self width - 2) / 2;
			layoutPolicy: TableLayout new;
			layoutInset: 1@5;
			cellInset: 9@11;
			listCentering: #center;
 			listDirection: #leftToRight;
			cellPositioning: #topCenter.
	1 to: AltFreeCell numPiles do: [ :n | 
		| pile |
		pile := AltCardPileMorph new initialize;
					setBalloonText: 'Build up a suit in order from the Ace.'.
		pilesArea addMorph: pile ].
	piles := pilesArea submorphs.
	self addMorph: pilesArea! !

!SWEKlondike methodsFor: 'auto-move' stamp: 'ChristopherMiller 2/23/2013 19:33'!
maxFaceNumberToMoveToPiles
	"Answer 1 plus the minimum Card faceNumber"

	| min |

	min := 13. "big"
	self piles 
		do: [ :pile |
			pile isEmpty 
			ifTrue: [ min := min min: 0 ] 
			ifFalse: [ min := min min: (pile submorphs at: 1) faceNumber ]
		    ].
	^min + 1! !

!SWEKlondike methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:33'!
newGame
	"Empty out my Card Containers, shuffle and re-deal."
	
	self removeWinnerText;
		makeUndoStackEmpty;
	    "Note: (self deck shuffle) is done in (self dealCards)."
		dealCards! !

!SWEKlondike methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
pickLegalFor: aCard 
	"Answer true if picking up a card from its container is a legal move."
	| cardContainer |
	cardContainer := aCard owner.
	self moveStart: cardContainer. "for Undo"
	(cardContainer isKindOf: AltCardContainerMorph)
		ifFalse: [^ false].
	cardContainer class = AltCardDiscardMorph
		"OK if top card"
		ifTrue: [ ^aCard == (cardContainer submorphs at: 1)].
	cardContainer class = AltCardPileMorph
		ifTrue: [^ false].
	cardContainer class = AltCardDeckMorph
		ifTrue: [^ false].
	cardContainer class = AltCardColumnMorph
		ifTrue: [^ self pickLegalFor: aCard fromColumn: cardContainer ].
	Error signal: 'Internal logic error'! !

!SWEKlondike methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
pickLegalFor: aCard fromColumn: aColumn
	"Answer true if card is on top or is part of a 'run'."

	^(self runStartsFrom: aCard in: aColumn) ! !

!SWEKlondike methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
pile: pileMorph wantsDropOf: aCardOrColumnMorph 
	"Answer true if Pile is empty and Card is Ace or if  
	card is same suit and one higher in face value."
	
	| card |
	card := aCardOrColumnMorph class = AltCardMorph
				ifTrue: [aCardOrColumnMorph]
				ifFalse: [aCardOrColumnMorph singletonCard].
	card isNil
		ifTrue: [ ^nil ]
		ifFalse: [ pileMorph isEmpty
			ifTrue: [ ^ card face == #Ace ]
			ifFalse: [ | topCard | 
			topCard := pileMorph topCard.
			^ (topCard hasFaceOneLessThan: card)  & (topCard hasSameSuitAs: card)]]! !

!SWEKlondike methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
piles
	"Private -- Answer my piles"

	^piles! !

!SWEKlondike methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:33'!
quit
	"@@ToDo: popup dialog Do you REALLY want to Quit?"
	
	super quit! !

!SWEKlondike methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:33'!
reDealCards
	"Deal cards onto the table withot shuffling (for game restart)."

	self deck makeEmpty; addCards: cardDeck.
	self removeCards.
	1 to: Klondike numColumns
		do: [ :colIndex |  | column |
			column := self columns at: colIndex.
			1 to: (colIndex - 1) 
				 do: [ :ignore | 
					column add: deck takeTopCard makeFaceDown ].
			column add: deck takeTopCard makeFaceUp.
		 ]! !

!SWEKlondike methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:33'!
removeCards
	"Empty out my Card Containers."
	
	discards makeEmpty.
	self piles do: [ :container | container makeEmpty ].
	self columns do: [ :container | container makeEmpty ]! !

!SWEKlondike methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:33'!
restoreGame
	"Put up a dialog to reaf game state from a file.
	See #restoreGameFromDataStream:"

	self restoreGame: 'klo' userMessage: 'Select a Klondike Game File...'! !

!SWEKlondike methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:33'!
restoreGameFromDataStream: aDataStream
	"Given an open DataStream, read and restore game state."
	
	"Game state consists of information to reconsitiute the cards, deck, discards, piles, and columns."
	"NYI: undo stack save/restore."
	| stateArray |
	stateArray := aDataStream next.
	(stateArray at: 1) == #Klondike
	ifFalse: [ Error signal: 'Not a saved Klondike Game!!' ] "FIXME: popup"
	ifTrue: [
		| columnInfo pileInfo deckInfo discardInfo undoInfo |
		cardDeck := AltCardDeck new initializeFromArray: (stateArray at: 3).
		columnInfo := stateArray at: 5.
		pileInfo     := stateArray at: 7.
		deckInfo    := stateArray at: 9.
		discardInfo := stateArray at: 11.
		undoInfo    := stateArray at: 13.
		
		1 to: (Klondike numColumns) 
			do: [ :idx |
				(self columns at: idx) restoreFromArray: (columnInfo at: idx) 
									givenDeck: self cardDeck ].
			
		1 to: (Klondike numPiles) 
			do: [ :idx |
				(self piles at: idx) restoreFromArray: (pileInfo at: idx) 
								  givenDeck: self cardDeck ].
		
		deck restoreFromArray: deckInfo givenDeck: self cardDeck.
		
		discards restoreFromArray: discardInfo givenDeck: self cardDeck.
			
		undoInfo isNil
		ifTrue: [ self makeUndoStackEmpty ]
		ifFalse: [ Error signal: 'FIXME: Undo save/restore NYI' ].
	]! !

!SWEKlondike methodsFor: 'testing' stamp: 'ChristopherMiller 2/23/2013 19:33'!
runStartsFrom: aCard in: aColumn
	"A Card starts a run if all cards above it are in ascending face and
	alternating color."

	| cardIndex cards cardUnder |
	cards := aColumn submorphs.
	cardIndex := cards indexOf: aCard.
	(cardIndex isNil) ifTrue: [ ^false ].
	cardUnder := aCard.
	[cardIndex > 1] 
		whileTrue: [ 
			| cardAbove |
			cardIndex := cardIndex - 1.
			cardAbove := cards at: cardIndex.
			(cardUnder hasDifferentColorThan: cardAbove)
			& (cardAbove hasFaceOneLessThan: cardUnder)
				ifFalse: [ ^false ].
			cardUnder := cardAbove.
		].
	"cardIndex = 1"
	^true! !

!SWEKlondike methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:33'!
sameGame
	"Empty out my Card Containers and deal the original deal."
	
	self cardDeck startOver.
	self removeWinnerText;
		makeUndoStackEmpty;
	    "Note: (self deck shuffle) is done in (self dealCards)."
		reDealCards! !

!SWEKlondike methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:33'!
saveGame
	"Show a file dialog to the user to save a game.
	See #saveGameToDataStream: for actual data save."
	
	| sillyDialog |
	self gameIsWon
	ifTrue: [  
		sillyDialog := DenyDialogWindow new.
		sillyDialog text: ('Sorry, I can''t save a game you finished playing' asText).
		self openModal: sillyDialog]
	ifFalse: [self saveGame: 'klo' userMessage: 'Enter File Name to save Klondike Game...']! !

!SWEKlondike methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:33'!
saveGameToDataStream: aDataStream
	"Given an open DataStream, save the current game state."
	
	"Game state consists of information to reconsitiute the cardDeck, deck, spares, discards, and columns."
	"NYI: undo stack save/restore."
	aDataStream nextPut: { #Klondike. 
							#CardDeck. "NB: AltCardDeck"
							self cardDeck saveStateToArray.
							#Columns.
							self columns 
								collect: [ :container | container saveStateToArray].
							#Piles.
							self piles 
								collect: [ :container | container saveStateToArray].
							#Deck. "NB: AltCardDeckMorph"
							self deck saveStateToArray.
							#Discards.
							self discard saveStateToArray.							
							#Undo.
							nil. "FIXME"
						}! !

AltCardTableMorph subclass: #SWEPyramid
	instanceVariableNames: 'deck spares piles columns'
	classVariableNames: 'NumColumns NumPiles NumSpares'
	poolDictionaries: ''
	category: '_SWE_OOAD'!

SWEPyramid class
	instanceVariableNames: 'NumPiles NumSpares NumColumns'!

SWEPyramid class
	instanceVariableNames: 'NumPiles NumSpares NumColumns'!

!SWEPyramid class methodsFor: 'parts bin' stamp: 'ChristopherMiller 2/23/2013 19:35'!
descriptionForPartsBin
	^ self
		partName: 'AltFreeCell'
		categories: #('Games' )
		documentation: 'Another FreeCell solitaire card game'! !

!SWEPyramid class methodsFor: 'class initialization' stamp: 'ChristopherMiller 2/23/2013 19:35'!
includeInNewMorphMenu
	^ true! !

!SWEPyramid class methodsFor: 'class initialization' stamp: 'ChristopherMiller 2/23/2013 19:35'!
initialize

	NumColumns := 8.
	NumSpares := 4.
	NumPiles := 4.
	^self! !

!SWEPyramid class methodsFor: 'instance creation' stamp: 'ChristopherMiller 2/23/2013 19:35'!
newGame
	"Answer a new instance"
	"AltFreeCell newGame openInWorld."
	^ self new initialize "openInWindowLabeled: 'FreeCell' inWorld: self currentWorld"! !

!SWEPyramid class methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
numColumns
	^NumColumns! !

!SWEPyramid class methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
numPiles
	^NumPiles! !

!SWEPyramid class methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
numSpares
	^NumSpares! !

!SWEPyramid methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
column: columnMorph wantsDropOf: aCardOrColumnMorph 
	"Answer true if Column is empty or if (lowest)
	card is opposit color and one higher in face value."
	
	aCardOrColumnMorph isAltCardMorph
		ifTrue: [| topCard | 
			topCard := columnMorph topCard.
			topCard isNil
			ifTrue: [^true]
			ifFalse: [ ^ (aCardOrColumnMorph hasFaceOneLessThan: topCard)
					& (topCard hasDifferentColorThan: aCardOrColumnMorph)]]
		ifFalse: [
			Error signal: '@@FIXME: implement for column drops!!'.
			^ false]! !

!SWEPyramid methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
columns
	"Private -- Answer my columns"

	^columns! !

!SWEPyramid methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
container: aContainerMorph wantsDropOf: aCardOrColumnMorph 
	"Answer true if legal drop by FreeCell game rules"
	| dropObjectClass |
	dropObjectClass := aCardOrColumnMorph class.
	(dropObjectClass = AltCardColumnMorph) | (dropObjectClass = AltCardMorph)
	ifFalse: [ ^false ]
	ifTrue: [ aContainerMorph class = AltCardSpareMorph
		ifTrue: [^ self spare: aContainerMorph wantsDropOf: aCardOrColumnMorph]
		ifFalse: [aContainerMorph class = AltCardPileMorph
				ifTrue: [^ self pile: aContainerMorph wantsDropOf: aCardOrColumnMorph]
				ifFalse: [aContainerMorph class = AltCardColumnMorph
						ifTrue: [^ self column: aContainerMorph wantsDropOf: aCardOrColumnMorph]
						ifFalse: [^ false]]]]! !

!SWEPyramid methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:35'!
dealCards
	"Deal all cards onto the table by adding them to Card Columns."

	self deck shuffle; shuffle; shuffle; shuffle; shuffle; shuffle; shuffle.
	0 to: 51
		do: [ :n |  
			(columns at: (n rem: AltFreeCell numColumns) + 1) add: (deck dealACard) ]
			! !

!SWEPyramid methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
deck
	"Private -- Answer my deck"

	^deck! !

!SWEPyramid methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
dropComplete
	"Check to see if any cards can be auto-moved to piles.
	Keep moving cards up if legal unless the lowest face value
	of a Card in any Pile is significantly less than the others
	as noted by #maxFaceNumberToMoveToPiles."
	
	self inCardMove 
		"already doing this"
		ifTrue: [ ^false ]
		ifFalse: 
			[ | goAgain |
				self inCardMove: true.
			 	goAgain := true.
				[ goAgain ] whileTrue:
					[ | maxFaceToMove |
				      goAgain := false.
					 maxFaceToMove := self maxFaceNumberToMoveToPiles.
					 self spares 
						do: [ :container |
							 (self possiblyAnimateCardsFrom: container 
								   ifFaceLessOrEqualTo: maxFaceToMove)
							 ifTrue: [ goAgain := true. ].
						].
			  		 self columns
						do: [ :container |
							  (self possiblyAnimateCardsFrom: container 
								   ifFaceLessOrEqualTo: maxFaceToMove) 
							  ifTrue: [ goAgain := true. ].
						]
				].
				self inCardMove: false.
				self gameIsWon
				ifTrue: [ self showWinnerText ]
			] ! !

!SWEPyramid methodsFor: 'testing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
gameIsWon
	"Answer true if all cards are in the piles"

	^(self piles inject: 0
			    into: [ :count :pile | count + (pile submorphs size) ]) == 52! !

!SWEPyramid methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:35'!
help

	self helpWithLabel: 'FreeCell Rules' translated
! !

!SWEPyramid methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
helpText
	^ '
	All cards are dealt into 8 columns.
 
	Cards in columns build up by alternating color and increasing sequence.

	You can move such "runs" up to the number of free cells plus 1.

	Free cells are the spares and columns which contain no cards.

	There are 4 spares. Each spare can hold a single card.

	The piles build up from the aces in the same suit.' translated! !

!SWEPyramid methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:35'!
initialize
	"Setup my layout.  Across the top are 4 Spares and 4 Piles,
	then a block of 8 Columns, then a Button area."

	super initialize.
	deck := AltCardDeck new initialize.
	self name: 'FreeCell';
		layoutPolicy: TableLayout new;
		listDirection: #leftToRight;
 		wrapDirection: #topToBottom;
		listCentering: #topLeft;
		makeButtonArea;
		makeColumnArea;
		makePilesArea;
		makeSparesArea.
	self dealCards;
		yourself! !

!SWEPyramid methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:35'!
makeColumnArea
	"I make new Column Area as a submorph"

	| columnArea |
	columnArea := 
		Morph new initialize; 
			color: Color transparent;
			"height: 460;"
			width: (self width);
			layoutPolicy: TableLayout new;
			layoutInset: 14;
 			listDirection: #leftToRight;
			listCentering: #justified;
			cellPositioning: #topCenter;
			vResizing: #spaceFill.
	1 to: AltFreeCell numColumns 
		do: [ :n |
			| colMorph |
			colMorph := AltCardColumnMorph new initialize.
			colMorph height: 400;
					  borderColor: Color transparent.
			columnArea addMorph: colMorph ].
	columns := columnArea submorphs.
	self addMorph: columnArea! !

!SWEPyramid methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:35'!
makePilesArea
	"I make new Piles Area as a submorph"

	| pilesArea |
	pilesArea := 
		Morph new initialize; 
			color: Color transparent;
			height: 95;
			width: (self width - 2) / 2;
			layoutPolicy: TableLayout new;
			layoutInset: 1@5;
			cellInset: 9@11;
			listCentering: #center;
 			listDirection: #leftToRight;
			cellPositioning: #topCenter.
	1 to: AltFreeCell numPiles do: [ :n | 
		| pile |
		pile := AltCardPileMorph new initialize;
					setBalloonText: 'Build up a suit in order from the Ace.'.
		pilesArea addMorph: pile ].
	piles := pilesArea submorphs.
	self addMorph: pilesArea! !

!SWEPyramid methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:35'!
makeSparesArea
	"I make new Piles Area as a submorph"

	| sparesArea |
	sparesArea := 
		Morph new initialize; 
			color: Color transparent;
			height: 95;
			width: (self width - 2) / 2;
			layoutPolicy: TableLayout new;
			layoutInset: 1@5;
			cellInset: 9@11;
			listCentering: #center;
 			listDirection: #leftToRight;
			cellPositioning: #topCenter.
	1 to: AltFreeCell numSpares 
		do: [ :n | 
			| spare |
			spare := AltCardSpareMorph new initialize; 
				setBalloonText: 'A Spare can hold 1 card'.
			sparesArea addMorph: spare ].
	spares := sparesArea submorphs.
	self addMorph: sparesArea! !

!SWEPyramid methodsFor: 'testing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
maxFaceNumberToMoveToPiles
	"Answer N plus the minimum Card faceNumber"

	| min |

	min := 13. "big"
	self piles 
		do: [ :pile |
			pile isEmpty 
			ifTrue: [ min := min min: 0 ] 
			ifFalse: [ min := min min: (pile submorphs at: 1) faceNumber ]
		    ].
	"N is 2"
	^min + 2! !

!SWEPyramid methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:35'!
newGame
	"Empty out my Card Containers, shuffle and re-deal."
	
	self removeWinnerText;
		makeUndoStackEmpty;
		removeCards;
	    "Note: (self deck shuffle) is done in (self dealCards)."
		dealCards! !

!SWEPyramid methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
numberOfFreeCells
	"Answer the number of empty Spares and Columns."

	| numFree |
	numFree := 0.
	self spares   do: [ :spare | spare isEmpty ifTrue: [numFree := numFree + 1]].
	self columns do: [ :spare | spare isEmpty ifTrue: [numFree := numFree + 1]].
	^numFree! !

!SWEPyramid methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
pickLegalFor: aCard 
	"Answer true if picking up a card from its container is a legal move."
	| cardContainer |
	cardContainer := aCard owner.
	self moveStart: cardContainer. "for Undo"
	(cardContainer isKindOf: AltCardContainerMorph)
		ifFalse: [^ false].
	cardContainer class = AltCardSpareMorph
		ifTrue: [^ true].
	cardContainer class = AltCardPileMorph
		ifTrue: [^ false].
	"else is a Column"
	^ self pickLegalFor: aCard fromColumn: cardContainer! !

!SWEPyramid methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
pickLegalFor: aCard fromColumn: aColumn
	"Answer true if card is on top or is part of a 'run' and the size of
	the run is less than the number of free cells."

	(self runStartsFrom: aCard in: aColumn) ifFalse: [ ^false ].
	^(self numberOfFreeCells + 1) >= (aColumn numberOfCardsAbove: aCard)! !

!SWEPyramid methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
pile: pileMorph wantsDropOf: aCardOrColumnMorph 
	"Answer true if Pile is empty and Card is Ace or if  
	card is same suit and one higher in face value."
	
	| card |
	card := aCardOrColumnMorph class = AltCardMorph
				ifTrue: [aCardOrColumnMorph]
				ifFalse: [aCardOrColumnMorph singletonCard].
	card isNil
		ifTrue: [ ^nil ]
		ifFalse: [ pileMorph isEmpty
			ifTrue: [ ^ card face == #Ace ]
			ifFalse: [ | topCard | 
			topCard := pileMorph topCard.
			^ (topCard hasFaceOneLessThan: card)  & (topCard hasSameSuitAs: card)]]! !

!SWEPyramid methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
piles
	"Private -- Answer my piles"

	^piles! !

!SWEPyramid methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:35'!
quit
	"@@ToDo: popup dialog Do you REALLY want to Quit?"
	
	super quit! !

!SWEPyramid methodsFor: 'initialization' stamp: 'ChristopherMiller 2/23/2013 19:35'!
reDealCards
	"Deal all cards onto the table withot shuffling (for game replay)."

	"Assumes cards have been removed first."
	0 to: 51
		do: [ :n |  
			(columns at: (n rem: AltFreeCell numColumns) + 1) add: (deck dealACard) ]
			! !

!SWEPyramid methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:35'!
removeCards
	"Empty out my Card Containers."
	
	self piles do: [ :container | container makeEmpty ].
	self spares do: [ :container | container makeEmpty ].
	self columns do: [ :container | container makeEmpty ]! !

!SWEPyramid methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:35'!
restoreGame
	"Put up a dialog to reaf game state from a file.
	See #restoreGameFromDataStream:"

	self restoreGame: 'fre' userMessage: 'Select a FreeCell Game File...'! !

!SWEPyramid methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:35'!
restoreGameFromDataStream: aDataStream
	"Given an open DataStream, read and restore game state."
	
	"Game state consists of information to reconsitiute the deck, spares, piles, and columns."
	"NYI: undo stack save/restore."
	| stateArray |
	stateArray := aDataStream next.
	(stateArray at: 1) == #AltFreeCell
	ifFalse: [ Error signal: 'Not a saved FreeCell Game!!' ]
	ifTrue: [
		| columnInfo pileInfo spareInfo undoInfo |
		deck := AltCardDeck new initializeFromArray: (stateArray at: 3).
		columnInfo := stateArray at: 5.
		pileInfo     := stateArray at: 7.
		spareInfo   := stateArray at: 9.
		undoInfo    := stateArray at: 11.
		
		1 to: (AltFreeCell numColumns) 
			do: [ :idx |
				(self columns at: idx) restoreFromArray: (columnInfo at: idx) givenDeck: self deck ].
		
		1 to: (AltFreeCell numSpares) 
			do: [ :idx |
				(self spares at: idx) restoreFromArray: (spareInfo at: idx) givenDeck: self deck ].
			
		1 to: (AltFreeCell numPiles) 
			do: [ :idx |
				(self piles at: idx) restoreFromArray: (pileInfo at: idx) givenDeck: self deck ].
			
		undoInfo isNil
		ifTrue: [ self makeUndoStackEmpty ]
		ifFalse: [ Error signal: 'FIXME: Undo save/restore NYI' ].
	]! !

!SWEPyramid methodsFor: 'actions' stamp: 'ChristopherMiller 2/23/2013 19:35'!
sameGame
	"Empty out my Card Containers and deal the original deal."
	
	self deck startOver.
	self removeWinnerText;
		makeUndoStackEmpty;
		removeCards;
	    "Note: (self deck shuffle) is done in (self dealCards)."
		reDealCards! !

!SWEPyramid methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:35'!
saveGame
	"Show a file dialog to the user to save a game.
	See #saveGameToDataStream: for actual data save."
	| sillyDialog |
	self gameIsWon
	ifTrue: [  
		sillyDialog := DenyDialogWindow new.
		sillyDialog text: ('Sorry, I can''t save a game you finished playing' asText).
		self openModal: sillyDialog]
	ifFalse: [self saveGame: 'fre' userMessage: 'Enter File Name to save FreeCell Game...']! !

!SWEPyramid methodsFor: 'fileIn/out' stamp: 'ChristopherMiller 2/23/2013 19:35'!
saveGameToDataStream: aDataStream
	"Given an open DataStream, save the current game state."
	
	"Game state consists of information to reconsitiute the deck, spares, piles, and columns."
	"NYI: undo stack save/restore."
	| someData |
	someData := 	 { #AltFreeCell. 
							#CardDeck.
							self deck saveStateToArray.
							#Columns.
							self columns 
								collect: [ :container | container saveStateToArray].
							#Piles.
							self piles 
								collect: [ :container | container saveStateToArray].
							#Spares.
							self spares 
								collect: [ :container | container saveStateToArray].
							#Undo.
							nil. "FIXME"
						}.
	aDataStream nextPut: someData.
! !

!SWEPyramid methodsFor: 'dropping/grabbing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
spare: spareMorph wantsDropOf: aCardOrColumnMorph 
	"Answer true if Spare is empty and Card is singleton."
	| card |
	card := aCardOrColumnMorph class = AltCardMorph
				ifTrue: [aCardOrColumnMorph]
				ifFalse: [aCardOrColumnMorph singletonCard].
	spareMorph isEmpty
		ifTrue: [^ card isNil not]
		ifFalse: [^ false]! !

!SWEPyramid methodsFor: 'accessing' stamp: 'ChristopherMiller 2/23/2013 19:35'!
spares
	"Private -- Answer my spares"

	^spares! !
SWEFreeCell initialize!
SWEKlondike initialize!
SWEPyramid initialize!
